@@ -1,6 +1,8 @@
 use core::fmt;
 use std::collections::HashMap;
 
+use crate::expander::determine_tag_method;
+use crate::expander::TagMethod;
 use crate::internals::symbol::*;
 use crate::internals::Ctxt;
 use crate::internals::Mode;
@@ -11,6 +13,23 @@ use syn::parse::Parse;
 use syn::spanned::Spanned;
 use syn::Ident;
 
+/// The value and method to encode / decode a tag.
+#[derive(Clone, Copy)]
+pub(crate) struct EnumTag<'a> {
+    pub(crate) value: &'a syn::Expr,
+    pub(crate) method: Option<TagMethod>,
+}
+
+#[derive(Clone, Copy)]
+pub(crate) enum EnumTagging<'a> {
+    /// The type is internally tagged by the field given by the expression.
+    Internal { tag: EnumTag<'a> },
+    Adjacent {
+        tag: EnumTag<'a>,
+        content: &'a syn::Expr,
+    },
+}
+
 /// The kind of tag to use.
 #[derive(Debug, Clone, Copy)]
 pub enum DefaultTag {
@@ -18,14 +37,8 @@ pub enum DefaultTag {
     Name,
 }
 
-impl Default for DefaultTag {
-    fn default() -> Self {
-        Self::Index
-    }
-}
-
 /// If the type is tagged or not.
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum Packing {
     Tagged,
     Packed,
@@ -52,37 +65,74 @@ impl Default for Packing {
 struct InnerTypeAttr {
     /// `#[musli(crate = <path>)]`.
     krate: Option<(Span, syn::ExprPath)>,
-    /// `#[musli(tag_type)]`.
+    /// `#[musli(name_type)]`.
     tag_type: Option<(Span, syn::Type)>,
-    /// `#[musli(default_variant_tag = "..")]`.
-    default_variant_tag: DefaultTag,
-    /// `#[musli(default_field_tag = "..")]`.
-    default_field_tag: DefaultTag,
+    /// `#[musli(default_variant_name = "..")]`.
+    default_variant_name: Option<DefaultTag>,
+    /// `#[musli(default_field_name = "..")]`.
+    default_field_name: Option<DefaultTag>,
+    /// If `#[musli(tag = <expr>)]` is specified.
+    tag: Option<(Span, syn::Expr)>,
+    /// If `#[musli(content = <expr>)]` is specified.
+    content: Option<(Span, syn::Expr)>,
     /// `#[musli(packed)]` or `#[musli(transparent)]`.
     packing: Option<(Span, Packing)>,
 }
 
 impl InnerTypeAttr {
-    /// Update packing of type.
-    fn set_packing(&mut self, cx: &Ctxt, span: Span, packing: Packing) {
-        if let Some((_, existing)) = self.packing {
+    /// Update how an enum is tagged.
+    fn set_tag(&mut self, cx: &Ctxt, span: Span, expr: syn::Expr) {
+        if self.tag.is_some() {
             cx.error_span(
                 span,
-                format!(
-                    "#[{}({})] cannot be combined with #[{}({})]",
-                    ATTR, packing, ATTR, existing
-                ),
+                format_args!("#[{ATTR}({TAG})] may only be specified once",),
             );
+
+            return;
+        }
+
+        self.tag = Some((span, expr));
+    }
+
+    /// Update how an enum is tagged.
+    fn set_content(&mut self, cx: &Ctxt, span: Span, expr: syn::Expr) {
+        if self.content.is_some() {
+            cx.error_span(
+                span,
+                format_args!("#[{ATTR}({TAG})] may only be specified once",),
+            );
+
+            return;
+        }
+
+        self.content = Some((span, expr));
+    }
+
+    /// Update packing of type.
+    fn set_packing(&mut self, cx: &Ctxt, span: Span, packing: Packing) -> bool {
+        if let Some((_, existing)) = &self.packing {
+            if *existing != packing {
+                cx.error_span(
+                    span,
+                    format_args!(
+                        "#[{}({})] cannot be combined with #[{}({})]",
+                        ATTR, packing, ATTR, existing
+                    ),
+                );
+
+                return false;
+            }
         }
 
         self.packing = Some((span, packing));
+        true
     }
 
     fn set_crate(&mut self, cx: &Ctxt, span: Span, path: syn::ExprPath) {
         if let Some((span, _)) = self.krate {
             cx.error_span(
                 span,
-                format!("#[{}({})] cannot be used multiple times", ATTR, CRATE),
+                format_args!("#[{}({})] cannot be used multiple times", ATTR, CRATE),
             );
         }
 
@@ -99,29 +149,63 @@ pub(crate) struct TypeAttr {
 
 impl TypeAttr {
     /// Indicates the packing state of the type.
-    pub(crate) fn packing(&self, mode: Mode<'_>) -> Option<(Span, Packing)> {
+    pub(crate) fn packing_span(&self, mode: Mode<'_>) -> Option<(Span, Packing)> {
         mode.ident
-            .and_then(|m| self.modes.get(m))
-            .and_then(|a| a.packing)
-            .or_else(|| self.root.packing)
+            .and_then(|m| self.modes.get(m)?.packing)
+            .or(self.root.packing)
+    }
+
+    fn tag(&self, mode: Mode<'_>) -> Option<&(Span, syn::Expr)> {
+        mode.ident
+            .and_then(|m| self.modes.get(m)?.tag.as_ref())
+            .or(self.root.tag.as_ref())
+    }
+
+    fn content(&self, mode: Mode<'_>) -> Option<&(Span, syn::Expr)> {
+        mode.ident
+            .and_then(|m| self.modes.get(m)?.content.as_ref())
+            .or(self.root.content.as_ref())
+    }
+
+    pub(crate) fn enum_tagging_span(&self, mode: Mode<'_>) -> Option<Span> {
+        let tag = self.tag(mode);
+        let content = self.content(mode);
+        Some(tag.or(content)?.0)
+    }
+
+    /// Indicates the state of enum tagging.
+    pub(crate) fn enum_tagging(&self, mode: Mode<'_>) -> Option<EnumTagging<'_>> {
+        let tag = self.tag(mode);
+        let (_, tag) = tag?;
+
+        let tag_method = determine_tag_method(tag);
+        let tag = EnumTag {
+            value: tag,
+            method: tag_method,
+        };
+
+        match self.content(mode) {
+            Some((_, content)) => Some(EnumTagging::Adjacent { tag, content }),
+            _ => Some(EnumTagging::Internal { tag }),
+        }
     }
 
     /// Indicates the packing state of the type.
-    pub(crate) fn packing_or_default(&self, mode: Mode<'_>) -> Packing {
-        self.packing(mode).map(|p| p.1).unwrap_or_default()
+    pub(crate) fn packing(&self, mode: Mode<'_>) -> Option<Packing> {
+        Some(self.packing_span(mode)?.1)
     }
 
     /// Default field tag.
-    pub(crate) fn default_field_tag(&self, mode: Mode<'_>) -> DefaultTag {
+    pub(crate) fn default_field_name(&self, mode: Mode<'_>) -> Option<DefaultTag> {
         mode.ident
-            .and_then(|m| Some(self.modes.get(m)?.default_field_tag))
-            .unwrap_or(self.root.default_field_tag)
+            .and_then(|m| Some(self.modes.get(m)?.default_field_name))
+            .unwrap_or(self.root.default_field_name)
     }
 
-    pub(crate) fn default_variant_tag(&self, mode: Mode<'_>) -> DefaultTag {
+    pub(crate) fn default_variant_name(&self, mode: Mode<'_>) -> Option<DefaultTag> {
         mode.ident
-            .and_then(|m| Some(self.modes.get(m)?.default_variant_tag))
-            .unwrap_or(self.root.default_variant_tag)
+            .and_then(|m| Some(self.modes.get(m)?.default_variant_name))
+            .unwrap_or(self.root.default_variant_name)
     }
 
     /// Get the tag type of the type.
@@ -166,7 +250,7 @@ impl InnerFieldAttr {
         if self.encode_path.is_some() {
             cx.error_spanned_by(
                 encode_path,
-                format!("#[{}] multiple encode methods specified", ATTR),
+                format_args!("#[{}] multiple encode methods specified", ATTR),
             );
         } else {
             self.encode_path = Some((span, encode_path));
@@ -177,7 +261,7 @@ impl InnerFieldAttr {
         if self.decode_path.is_some() {
             cx.error_spanned_by(
                 decode_path,
-                format!("#[{}] multiple decode methods specified", ATTR),
+                format_args!("#[{}] multiple decode methods specified", ATTR),
             );
         } else {
             self.decode_path = Some((span, decode_path));
@@ -188,7 +272,7 @@ impl InnerFieldAttr {
         if self.skip_encoding_if.is_some() {
             cx.error_spanned_by(
                 skip_encoding_if,
-                format!("#[{}] multiple skip_encoding_if methods specified", ATTR),
+                format_args!("#[{}] multiple skip_encoding_if methods specified", ATTR),
             );
         } else {
             self.skip_encoding_if = Some((span, skip_encoding_if));
@@ -198,7 +282,7 @@ impl InnerFieldAttr {
 
 #[derive(Default)]
 struct InternalVariantAttr {
-    /// `#[musli(tag_type)]`.
+    /// `#[musli(name_type)]`.
     tag_type: Option<(Span, syn::Type)>,
     /// Rename a field to the given expression.
     rename: Option<(Span, syn::Expr)>,
@@ -206,17 +290,17 @@ struct InternalVariantAttr {
     packing: Option<(Span, Packing)>,
     /// `#[musli(default)]`.
     default: Option<Span>,
-    /// `#[musli(default_field_tag = "..")]`.
-    default_field_tag: Option<DefaultTag>,
+    /// `#[musli(default_field_name = "..")]`.
+    default_field_name: Option<DefaultTag>,
 }
 
 impl InternalVariantAttr {
     /// Update packing of type.
     fn set_packing(&mut self, cx: &Ctxt, span: Span, packing: Packing) {
-        if let Some((_, existing)) = self.packing {
+        if let Some((_, existing)) = &self.packing {
             cx.error_span(
                 span,
-                format!(
+                format_args!(
                     "#[{}({})] cannot be combined with #[{}({})]",
                     ATTR, packing, ATTR, existing
                 ),
@@ -250,16 +334,19 @@ impl VariantAttr {
 
     /// Indicates if the tagged state of the variant is set.
     pub(crate) fn packing(&self, mode: Mode<'_>) -> Option<Packing> {
-        mode.ident
-            .and_then(|m| Some(self.modes.get(m)?.packing?.1))
-            .or_else(|| Some(self.root.packing?.1))
+        let packing = mode
+            .ident
+            .and_then(|m| self.modes.get(m)?.packing.as_ref())
+            .or(self.root.packing.as_ref())?;
+
+        Some(packing.1)
     }
 
     /// Default field tag.
-    pub(crate) fn default_field_tag(&self, mode: Mode<'_>) -> Option<DefaultTag> {
+    pub(crate) fn default_field_name(&self, mode: Mode<'_>) -> Option<DefaultTag> {
         mode.ident
-            .and_then(|m| self.modes.get(m)?.default_field_tag)
-            .or_else(|| self.root.default_field_tag)
+            .and_then(|m| self.modes.get(m)?.default_field_name)
+            .or_else(|| self.root.default_field_name)
     }
 
     /// Get the tag type of the type.
@@ -306,7 +393,7 @@ impl FieldAttr {
                 quote_spanned!(*span => #encode_path::<#mode_ident, _>),
             )
         } else {
-            let encode_path = mode.encode_t_encode();
+            let encode_path = mode.encode_t_encode(span);
             (span, encode_path)
         }
     }
@@ -325,7 +412,7 @@ impl FieldAttr {
                 quote_spanned!(*span => #decode_path::<#mode_ident, _>),
             )
         } else {
-            let decode_path = mode.decode_t_decode();
+            let decode_path = mode.decode_t_decode(span);
             (span, decode_path)
         }
     }
@@ -356,44 +443,60 @@ pub(crate) fn type_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> TypeAttr {
 
             for attribute in attributes.attributes {
                 match attribute {
+                    // parse #[musli(tag = <expr>)]
+                    Attribute::KeyValue(path, value) if path == TAG => {
+                        if let Some(expr) = value_as_expr(cx, TAG, value) {
+                            attr.set_tag(cx, path.span(), expr);
+                        }
+                    }
+                    // parse #[musli(content = <expr>)]
+                    Attribute::KeyValue(path, value) if path == CONTENT => {
+                        if let Some(expr) = value_as_expr(cx, CONTENT, value) {
+                            attr.set_content(cx, path.span(), expr);
+                        }
+                    }
                     // parse #[musli(crate = <path>)]
                     Attribute::KeyValue(path, value) if path == CRATE => {
                         if let Some(path) = value_as_path(cx, CRATE, value) {
                             attr.set_crate(cx, path.span(), path);
                         }
                     }
-                    // parse #[musli(tag_type = <type>)]
-                    Attribute::KeyValue(path, value) if path == TAG_TYPE => {
-                        if let Some(ty) = value_as_type(cx, TAG_TYPE, value) {
+                    // parse #[musli(name_type = <type>)]
+                    Attribute::KeyValue(path, value) if path == NAME_TYPE => {
+                        if let Some(ty) = value_as_type(cx, NAME_TYPE, value) {
                             attr.tag_type = Some((path.span(), ty));
                         }
                     }
-                    // parse #[musli(default_variant_tag = "..")]
-                    Attribute::KeyValue(path, expr) if path == DEFAULT_VARIANT_TAG => {
-                        if let Some(tag) = parse_value_string(cx, DEFAULT_VARIANT_TAG, expr) {
-                            attr.default_variant_tag = match tag.value().as_str() {
-                                "index" => DefaultTag::Index,
-                                "name" => DefaultTag::Name,
+                    // parse #[musli(default_variant_name = "..")]
+                    Attribute::KeyValue(path, expr) if path == DEFAULT_VARIANT_NAME => {
+                        if let Some(tag) = parse_value_string(cx, DEFAULT_VARIANT_NAME, expr) {
+                            attr.default_variant_name = match tag.value().as_str() {
+                                "index" => Some(DefaultTag::Index),
+                                "name" => Some(DefaultTag::Name),
                                 _ => {
                                     cx.error_spanned_by(
                                         tag,
-                                        format!("illegal #[{}({})] value", ATTR, TAG),
+                                        format_args!(
+                                            "illegal #[{ATTR}({DEFAULT_VARIANT_NAME})] value"
+                                        ),
                                     );
                                     continue;
                                 }
                             };
                         }
                     }
-                    // parse #[musli(default_field_tag = "..")]
-                    Attribute::KeyValue(path, expr) if path == DEFAULT_FIELD_TAG => {
-                        if let Some(tag) = parse_value_string(cx, DEFAULT_FIELD_TAG, expr) {
-                            attr.default_field_tag = match tag.value().as_str() {
-                                "index" => DefaultTag::Index,
-                                "name" => DefaultTag::Name,
+                    // parse #[musli(default_field_name = "..")]
+                    Attribute::KeyValue(path, expr) if path == DEFAULT_FIELD_NAME => {
+                        if let Some(tag) = parse_value_string(cx, DEFAULT_FIELD_NAME, expr) {
+                            attr.default_field_name = match tag.value().as_str() {
+                                "index" => Some(DefaultTag::Index),
+                                "name" => Some(DefaultTag::Name),
                                 _ => {
                                     cx.error_spanned_by(
                                         tag,
-                                        format!("illegal #[{}({})] value", ATTR, TAG),
+                                        format_args!(
+                                            "illegal #[{ATTR}({DEFAULT_FIELD_NAME})] value"
+                                        ),
                                     );
                                     continue;
                                 }
@@ -409,7 +512,11 @@ pub(crate) fn type_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> TypeAttr {
                         attr.set_packing(cx, path.span(), Packing::Transparent);
                     }
                     meta => {
-                        cx.error_span(meta.span(), format!("unsupported #[{}] attribute", ATTR));
+                        let path = format_path(meta.path());
+                        cx.error_span(
+                            meta.span(),
+                            format_args!("#[{ATTR}({path})] unsupported type attribute"),
+                        );
                     }
                 }
             }
@@ -463,11 +570,11 @@ pub(crate) fn field_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> FieldAttr {
                             attr.set_skip_encoding_if(cx, path.span(), path.clone());
                         }
                     }
-                    // parse #[musli(tag = <expr>)]
-                    Attribute::KeyValue(path, value) if path == TAG => {
+                    // parse #[musli(rename = <expr>)]
+                    Attribute::KeyValue(path, value) if path == RENAME => {
                         let span = path.span();
 
-                        if let Some(expr) = value_as_expr(cx, TAG, value) {
+                        if let Some(expr) = value_as_expr(cx, RENAME, value) {
                             if let Some((span, _)) = &attr.rename {
                                 cx.error_span(*span, "conflicting attribute");
                             } else {
@@ -480,14 +587,18 @@ pub(crate) fn field_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> FieldAttr {
                         if let Some(span) = attr.default {
                             cx.error_span(
                                 span,
-                                format!("#[{}({})] was previously defined here", ATTR, DEFAULT),
+                                format_args!("#[{ATTR}({DEFAULT})] was previously defined here"),
                             );
                         } else {
                             attr.default = Some(path.span());
                         }
                     }
                     meta => {
-                        cx.error_span(meta.span(), format!("unsupported #[{}] attribute", ATTR));
+                        let path = format_path(meta.path());
+                        cx.error_span(
+                            meta.span(),
+                            format_args!("#[{ATTR}({path})] unsupported field attribute"),
+                        );
                     }
                 }
             }
@@ -513,17 +624,17 @@ pub(crate) fn variant_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> VariantAttr
 
             for attribute in attributes.attributes {
                 match attribute {
-                    // parse #[musli(tag_type = <type>)]
-                    Attribute::KeyValue(path, value) if path == TAG_TYPE => {
-                        if let Some(ty) = value_as_type(cx, TAG_TYPE, value) {
+                    // parse #[musli(name_type = <type>)]
+                    Attribute::KeyValue(path, value) if path == NAME_TYPE => {
+                        if let Some(ty) = value_as_type(cx, NAME_TYPE, value) {
                             attr.tag_type = Some((path.span(), ty));
                         }
                     }
-                    // parse #[musli(tag = <expr>)]
-                    Attribute::KeyValue(path, value) if path == TAG => {
+                    // parse #[musli(rename = <expr>)]
+                    Attribute::KeyValue(path, value) if path == RENAME => {
                         let span = path.span();
 
-                        if let Some(expr) = value_as_expr(cx, TAG, value) {
+                        if let Some(expr) = value_as_expr(cx, RENAME, value) {
                             if let Some((span, _)) = &attr.rename {
                                 cx.error_span(*span, "conflicting attribute");
                             } else {
@@ -531,16 +642,18 @@ pub(crate) fn variant_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> VariantAttr
                             }
                         }
                     }
-                    // parse #[musli(default_field_tag = "..")]
-                    Attribute::KeyValue(path, expr) if path == DEFAULT_FIELD_TAG => {
-                        if let Some(tag) = parse_value_string(cx, DEFAULT_FIELD_TAG, expr) {
-                            attr.default_field_tag = Some(match tag.value().as_str() {
+                    // parse #[musli(default_field_name = "..")]
+                    Attribute::KeyValue(path, expr) if path == DEFAULT_FIELD_NAME => {
+                        if let Some(tag) = parse_value_string(cx, DEFAULT_FIELD_NAME, expr) {
+                            attr.default_field_name = Some(match tag.value().as_str() {
                                 "index" => DefaultTag::Index,
                                 "name" => DefaultTag::Name,
                                 _ => {
                                     cx.error_spanned_by(
                                         tag,
-                                        format!("illegal #[{}({})] value", ATTR, TAG),
+                                        format_args!(
+                                            "illegal #[{ATTR}({DEFAULT_FIELD_NAME})] value"
+                                        ),
                                     );
                                     continue;
                                 }
@@ -560,7 +673,11 @@ pub(crate) fn variant_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> VariantAttr
                         attr.default = Some(path.span());
                     }
                     meta => {
-                        cx.error_span(meta.span(), format!("unsupported #[{}] attribute", ATTR));
+                        let path = format_path(meta.path());
+                        cx.error_span(
+                            meta.span(),
+                            format_args!("#[{ATTR}({path})] unsupported variant attribute"),
+                        );
                     }
                 }
             }
@@ -577,7 +694,7 @@ fn value_as_path<'a>(cx: &Ctxt, attr: Symbol, value: AttributeValue) -> Option<s
         _ => {
             cx.error_span(
                 value.span(),
-                format!("#[{} = ...] should be a simple path", attr),
+                format_args!("#[{ATTR}({attr} = ..)] should be a simple path"),
             );
             None
         }
@@ -591,7 +708,7 @@ fn value_as_expr<'a>(cx: &Ctxt, attr: Symbol, value: AttributeValue) -> Option<s
         _ => {
             cx.error_span(
                 value.span(),
-                format!("#[{} = ...] should be a simple path", attr),
+                format_args!("#[{ATTR}({attr} = ..)] should be a simple path"),
             );
             None
         }
@@ -603,7 +720,10 @@ fn parse_value_string(cx: &Ctxt, attr: Symbol, value: AttributeValue) -> Option<
     match value {
         AttributeValue::Lit(syn::Lit::Str(string)) => Some(string),
         expr => {
-            cx.error_span(expr.span(), format!("#[{} = ...] should be a string", attr));
+            cx.error_span(
+                expr.span(),
+                format_args!("#[{} = ...] should be a string", attr),
+            );
             None
         }
     }
@@ -614,7 +734,10 @@ fn value_as_type(cx: &Ctxt, attr: Symbol, value: AttributeValue) -> Option<syn::
     match value {
         AttributeValue::Type(ty) => Some(ty),
         expr => {
-            cx.error_span(expr.span(), format!("#[{} = ...] should be a type", attr));
+            cx.error_span(
+                expr.span(),
+                format_args!("#[{} = ...] should be a type", attr),
+            );
             None
         }
     }
@@ -668,6 +791,16 @@ enum Attribute {
     KeyValue(syn::Path, AttributeValue),
 }
 
+impl Attribute {
+    /// Extract the path of the attribute.
+    fn path(&self) -> &syn::Path {
+        match self {
+            Attribute::Path(path) => path,
+            Attribute::KeyValue(path, _) => path,
+        }
+    }
+}
+
 impl Spanned for Attribute {
     fn span(&self) -> Span {
         match self {
@@ -706,16 +839,24 @@ impl Parse for TypeAttributes {
 
                         continue;
                     }
+                    // parse #[musli(tag = <expr>)]
+                    path if path == TAG => AttributeValue::Expr(content.parse()?),
+                    // parse #[musli(content = <expr>)]
+                    path if path == CONTENT => AttributeValue::Expr(content.parse()?),
                     // parse #[musli(crate = <path>)]
                     path if path == CRATE => AttributeValue::Path(content.parse()?),
-                    // parse #[musli(tag_type = <type>)]
-                    path if path == TAG_TYPE => AttributeValue::Type(content.parse()?),
-                    // parse #[musli(default_variant_tag = "..")]
-                    path if path == DEFAULT_VARIANT_TAG => AttributeValue::Lit(content.parse()?),
-                    // parse #[musli(default_field_tag = "..")]
-                    path if path == DEFAULT_FIELD_TAG => AttributeValue::Lit(content.parse()?),
+                    // parse #[musli(name_type = <type>)]
+                    path if path == NAME_TYPE => AttributeValue::Type(content.parse()?),
+                    // parse #[musli(default_variant_name = "..")]
+                    path if path == DEFAULT_VARIANT_NAME => AttributeValue::Lit(content.parse()?),
+                    // parse #[musli(default_field_name = "..")]
+                    path if path == DEFAULT_FIELD_NAME => AttributeValue::Lit(content.parse()?),
                     path => {
-                        return Err(syn::Error::new(path.span(), "unsupported attribute"));
+                        let p = format_path(path);
+                        return Err(syn::Error::new(
+                            path.span(),
+                            format_args!("#[{ATTR}({p})] unsupported type attribute"),
+                        ));
                     }
                 };
 
@@ -766,14 +907,18 @@ impl Parse for VariantAttributes {
 
                         continue;
                     }
-                    // parse #[musli(tag_type = <type>)]
-                    path if path == TAG_TYPE => AttributeValue::Type(content.parse()?),
-                    // parse #[musli(tag = <expr>)]
-                    path if path == TAG => AttributeValue::Expr(content.parse()?),
-                    // parse #[musli(default_field_tag = <expr>)]
-                    path if path == DEFAULT_FIELD_TAG => AttributeValue::Lit(content.parse()?),
+                    // parse #[musli(name_type = <type>)]
+                    path if path == NAME_TYPE => AttributeValue::Type(content.parse()?),
+                    // parse #[musli(rename = <expr>)]
+                    path if path == RENAME => AttributeValue::Expr(content.parse()?),
+                    // parse #[musli(default_field_name = <expr>)]
+                    path if path == DEFAULT_FIELD_NAME => AttributeValue::Lit(content.parse()?),
                     path => {
-                        return Err(syn::Error::new(path.span(), "unsupported attribute"));
+                        let p = format_path(path);
+                        return Err(syn::Error::new(
+                            path.span(),
+                            format_args!("#[{ATTR}({p})] unsupported variant attribute"),
+                        ));
                     }
                 };
 
@@ -828,10 +973,14 @@ impl Parse for FieldAttributes {
                     path if path == WITH => AttributeValue::Path(content.parse()?),
                     // parse #[musli(skip_encoding_if)]
                     path if path == SKIP_ENCODING_IF => AttributeValue::Path(content.parse()?),
-                    // parse #[musli(tag = <expr>)]
-                    path if path == TAG => AttributeValue::Expr(content.parse()?),
+                    // parse #[musli(rename = <expr>)]
+                    path if path == RENAME => AttributeValue::Expr(content.parse()?),
                     path => {
-                        return Err(syn::Error::new(path.span(), "unsupported attribute"));
+                        let p = format_path(path);
+                        return Err(syn::Error::new(
+                            path.span(),
+                            format_args!("#[{ATTR}({p})] unsupported field attribute"),
+                        ));
                     }
                 };
 
@@ -852,3 +1001,30 @@ impl Parse for FieldAttributes {
         })
     }
 }
+
+/// Format implementation of a path which ignores anything except the
+/// identifier components.
+fn format_path(path: &syn::Path) -> impl fmt::Display + '_ {
+    FormatPath { path }
+}
+
+struct FormatPath<'a> {
+    path: &'a syn::Path,
+}
+
+impl<'a> fmt::Display for FormatPath<'a> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let mut it = self.path.segments.iter();
+        let last = it.next_back();
+
+        for part in it {
+            write!(f, "{}::", part.ident)?;
+        }
+
+        if let Some(part) = last {
+            write!(f, "{}", part.ident)?;
+        }
+
+        Ok(())
+    }
+}
