@@ -1,8 +1,8 @@
 //! Helper for determining the mode we're currently in.
 
 use crate::internals::tokens::Tokens;
-use proc_macro2::TokenStream;
-use quote::{quote, ToTokens};
+use proc_macro2::{Span, TokenStream};
+use quote::{quote_spanned, ToTokens};
 
 #[derive(Clone, Copy)]
 pub(crate) enum ModePath<'a> {
@@ -32,21 +32,21 @@ pub(crate) struct Mode<'a> {
 
 impl<'a> Mode<'a> {
     /// Get the mode identifier.
-    pub(crate) fn mode_ident(&self) -> ModePath<'_> {
+    pub(crate) fn mode_ident(&self) -> ModePath<'a> {
         self.mode_path
     }
 
     /// Construct a typed encode call.
-    pub(crate) fn encode_t_encode(&self) -> TokenStream {
+    pub(crate) fn encode_t_encode(&self, span: Span) -> TokenStream {
         let moded_ident = &self.mode_path;
         let encode_t = &self.tokens.encode_t;
-        quote!(#encode_t::<#moded_ident>::encode)
+        quote_spanned!(span => #encode_t::<#moded_ident>::encode)
     }
 
     /// Construct a typed encode call.
-    pub(crate) fn decode_t_decode(&self) -> TokenStream {
+    pub(crate) fn decode_t_decode(&self, span: Span) -> TokenStream {
         let moded_ident = &self.mode_path;
         let decode_t = &self.tokens.decode_t;
-        quote!(#decode_t::<#moded_ident>::decode)
+        quote_spanned!(span => #decode_t::<#moded_ident>::decode)
     }
 }
